//
// Gaggle, simple example game implementation
// A sort-of P2P multiplayer game environment
// by Jarno van der Linden  jvan006@cs.auckland.ac.nz
//

#include "gaggle.h"

#include <list>
#include <iostream>
#include <sstream>

#include <cmath>

#include <minmax.h>
#include <conio.h>


#ifndef PI
#define PI 3.1415926535897932384
#endif
#ifndef M_PI
#define M_PI PI
#endif

// Path to where the model and texture files are located
#define FILEPATH "./"


//
// A fairly trivial example Home Player
//

class MyHomePlayer : public Gaggle::HomePlayer
{
	public:
		MyHomePlayer(const std::string& name, Gaggle::Model* model_in, Gaggle::Texture* texture_in);
		~MyHomePlayer();

		void Action(Gaggle::ActionCode code, float value = 0);
		void Text(const std::string& text);

		bool IsMoving(void) const;

	protected:
		void JoinedWorld(const Gaggle::WelcomePlayerMessage& msg);
		void LeftWorld(void);

		void WorldObjects(const Gaggle::ObjectsMessage& msg);
		void ObjectJoined(const Gaggle::ObjectJoinedMessage& msg);
		void ObjectLeft(const Gaggle::ObjectLeftMessage& msg);
		void ObjectState(const Gaggle::ObjectStateMessage& msg);

		const Gaggle::Texture* AskTexture(void);
		const Gaggle::Model* AskModel(void);

		void WorldTexture(const Gaggle::TextureMessage& msg);
		void WorldModel(const Gaggle::ModelMessage& msg);
		void WorldLayout(const Gaggle::WorldLayoutMessage& msg);

		void WorldTextMessage(const Gaggle::TextMessage& msg);
		void WorldScoreMessage(const Gaggle::ScoreMessage& msg);

		void WorldIntro(const Gaggle::WorldIntroMessage& msg);

		void ChangeWorld(const Gaggle::ChangeWorldMessage& msg);

	protected:
		// My model and texture
		Gaggle::Model* model;
		Gaggle::Texture* texture;

		// Keep a list of the textures in the current World.
		// A list may not the best way of storing the textures,
		// depending on how you render the World.
		// You should definitely not ask the World for a Texture
		// every time you need it. That would be too slow, and a
		// big load on the World's server. Get the texture once
		// and keep a local copy. Textures don't change.
		std::list<Gaggle::Texture*> textures;
		// Should have something similar for Models

		// We'll wait until we have the World Layout before we
		// start to play in the World
		bool havelayout;
};


MyHomePlayer::MyHomePlayer(const std::string& name, Gaggle::Model* model_in, Gaggle::Texture* texture_in)
	: Gaggle::HomePlayer(name), model(model_in), texture(texture_in), havelayout(false)
{
}

MyHomePlayer::~MyHomePlayer()
{
	for(std::list<Gaggle::Texture*>::iterator ti = textures.begin(); ti != textures.end(); ti++)
		delete *ti;
}


// Attempt to do an action
// The action request is sent to the World the player is
// currently in. The World should (eventually) send a
// message updating the Player position, orientation, etc.
// Note that this is only an action request. The World is
// free to alter the request, for example for collision
// response, limiting how fast the Player can move, or
// teleporting the Player to some other part of the World.
void MyHomePlayer::Action(Gaggle::ActionCode code, float value)
{
	std::cerr << "Action code " << code << " value " << value << std::endl;
	Gaggle::PlayerActionMessage msg;
	msg.Fill(msg, *this, code, value);
	SendToWorld(msg);
}


// Send a text message
// The text message is sent to the World the player is currently in
// What happens after that is up to the World
void MyHomePlayer::Text(const std::string& text)
{
	std::cerr << "Text " << text << std::endl;
	Gaggle::TextMessage msg;
	msg.Fill(msg, text);
	SendToWorld(msg);
}


// Return true if we're moving
bool MyHomePlayer::IsMoving(void) const
{
	Gaggle::Vector dpos, ddpos;
	GetDPos(dpos);
	GetDDPos(ddpos);
	float dheading = GetDHeading();
	if((dpos.x != 0) || (dpos.y != 0) || (dpos.z != 0) ||
	   (ddpos.x != 0) || (ddpos.y != 0) || (ddpos.z != 0) ||
		(dheading != 0))
		return true;
	return false;
}


// Player has joined a world. This is called as a result of receiving
// a welcome message from the World.
// Do any needed processing
void MyHomePlayer::JoinedWorld(const Gaggle::WelcomePlayerMessage& msg)
{
	std::cerr << "Player has joined world " << msg.name << ". UID set to " << msg.uid << std::endl;

	uid = msg.uid;

	Text("Hello everybody");
}

// The Player has left the World, either because the World has disconnected,
// or the Player has disconnected from the World
// Do any needed processing
void MyHomePlayer::LeftWorld(void)
{
	std::cerr << "Player has left world" << std::endl;
}


// The Player is informed about the objects in a World
// Do any needed processing
void MyHomePlayer::WorldObjects(const Gaggle::ObjectsMessage& msg)
{
	std::cerr << "Got informed about objects" << std::endl;
	std::cerr << "  " << msg.numobjects << " objects" << std::endl;
	// Do things like requesting any Models and Textures used by
	// the Objects we don't know about yet.
}


// Another Object has joined the same World as the Player is in
// Do any needed processing
void MyHomePlayer::ObjectJoined(const Gaggle::ObjectJoinedMessage& msg)
{
	std::cerr << "Player has been joined by another object with UID " << msg.uid << std::endl;
	// Do things like requesting any Models and Textures used by
	// the Object if we don't know about them yet. Do not ask for
	// the Model and Texture every time an object joins.
}


// Another Object in the same World as the Player has left
// Do any needed processing
void MyHomePlayer::ObjectLeft(const Gaggle::ObjectLeftMessage& msg)
{
	std::cerr << "Object with UID " << msg.uid << " has left the Player's world" << std::endl;

	// It's not a good idea to immediately get rid of the Object's
	// Model and Texture, as they may be reused for other Objects.
	// Think of missiles for example.
}


// The state of an Object is updated. That Object may be me as well
// Do any needed processing
void MyHomePlayer::ObjectState(const Gaggle::ObjectStateMessage& msg)
{
	std::cerr << "Object UID " << msg.uid << " state update" << std::endl;
	std::cerr << "  pos " << msg.pos.x << " " << msg.pos.y << " " << msg.pos.z << std::endl;
	std::cerr << "  dpos " << msg.dpos.x << " " << msg.dpos.y << " " << msg.dpos.z << std::endl;
	std::cerr << "  ddpos " << msg.ddpos.x << " " << msg.ddpos.y << " " << msg.ddpos.z << std::endl;

	if(msg.uid == uid)
	{
		std::cerr << "It's me!" << std::endl;
		// The World is the authority about what my state is
		SetState(msg);
	}
}


// The Player is asked about its texture
// Do any needed processing, possibly returning a texture
const Gaggle::Texture* MyHomePlayer::AskTexture(void)
{
	std::cerr << "Got asked about my texture" << std::endl;

	return texture;
}


// The Player is asked about its model
// Do any needed processing, possibly returning a model
const Gaggle::Model* MyHomePlayer::AskModel(void)
{
	std::cerr << "Got asked about my model" << std::endl;

	return model;
}


// The Player is informed about a texture
// Do any needed processing
void MyHomePlayer::WorldTexture(const Gaggle::TextureMessage& msg)
{
	std::cerr << "Got informed about texture " << msg.uid << std::endl;
	// Check to see if we already know about this texture
	for(std::list<Gaggle::Texture*>::iterator ti = textures.begin(); ti != textures.end(); ti++)
	{
		if((*ti)->GetUID() == msg.uid)
			return;
	}
	// Remember this texture
	Gaggle::Texture* worldtexture = new Gaggle::Texture(msg);
	textures.push_back(worldtexture);
}


// The Player is informed about a model
// Do any needed processing
void MyHomePlayer::WorldModel(const Gaggle::ModelMessage& msg)
{
	std::cerr << "Got informed about model " << msg.uid << std::endl;
	// Should be something like WorldTexture()
}


// The player has received world layout after a request
// Do any needed processing
void MyHomePlayer::WorldLayout(const Gaggle::WorldLayoutMessage& msg)
{
	std::cerr << "Got world layout for rect (" << msg.box[0][0] << ", " << msg.box[0][1] << ")";
	std::cerr << "->(" << msg.box[1][0] << ", " << msg.box[1][1] << ")" << std::endl;

	if(!havelayout)
	{
		// Now have some layout and are ready to play
		havelayout = true;
		Gaggle::PlayerReadyMessage readymsg;
		readymsg.Fill(readymsg);
		SendToWorld(readymsg);
	}
}


// The player has received a text message from the World
// Do any needed processing, such as displaying it on screen
void MyHomePlayer::WorldTextMessage(const Gaggle::TextMessage& msg)
{
	std::cerr << "Got text message:" << std::endl;
	std::cerr << "--" << std::endl;
	std::cerr << msg.text << std::endl;
	std::cerr << "--" << std::endl;
}


// The player has received a score message from the World,
// meant to show the player their current score, health, etc.
// Do any needed processing, such as displaying it on screen
void MyHomePlayer::WorldScoreMessage(const Gaggle::ScoreMessage& msg)
{
	std::cerr << "Got score message:" << std::endl;
	std::cerr << "--" << std::endl;
	std::cerr << msg.text << std::endl;
	std::cerr << "--" << std::endl;
}


// The player has received a world intro message from the World
// Do any needed processing
void MyHomePlayer::WorldIntro(const Gaggle::WorldIntroMessage& msg)
{
	std::cerr << "Got world intro name " << msg.name << std::endl;
	std::cerr << "  numsquares (" << msg.numsquares[0] << ", " << msg.numsquares[1] << ")" << std::endl;
	std::cerr << "  squaresize " << msg.squaresize << std::endl;
	std::cerr << "  my initial pos (" << msg.initpos.x << ", " << msg.initpos.y << ", " << msg.initpos.z << ")";
	std::cerr << "  heading " << msg.initheading << std::endl;

	// Ask about what the world looks like for 5 squares around me, for example
	Gaggle::AskWorldLayoutMessage asklayoutmsg;
	const int px = (int)floor(msg.initpos.x / msg.squaresize);
	const int py = (int)floor(msg.initpos.z / msg.squaresize);
	asklayoutmsg.Fill(asklayoutmsg, px-5, py-5, px+5, py+5);
	SendToWorld(asklayoutmsg);

	// Ask for existing objects
	Gaggle::AskObjectsMessage askobjectsmsg;
	askobjectsmsg.Fill(askobjectsmsg);
	SendToWorld(askobjectsmsg);
}


// The world is asking the player to switch to a different
// world. The new world is identified by a networkid.
void MyHomePlayer::ChangeWorld(const Gaggle::ChangeWorldMessage& msg)
{
	Leave();
	Join(msg.otherworld);
}


//
// A fairly trivial example Home World
//


// Keep some per-player info for all the Players in my Home World
// In this case, we keep track of the amount of ammo the Player has
class MyPlayer : public Gaggle::Player
{
	public:
		MyPlayer(const Gaggle::JoinPlayerMessage& msg, Gaggle::NetworkID networkid);
		~MyPlayer();

		int GetAmmo(void) const;
		void SetAmmo(int ammo);
		int AddAmmo(int ammo);
		int UseAmmo(int& use);

		std::string GetScoreText(void) const;
	protected:
		int ammo;
};

MyPlayer::MyPlayer(const Gaggle::JoinPlayerMessage& msg, Gaggle::NetworkID networkid)
	: Gaggle::Player(msg, networkid), ammo(0)
{
};

MyPlayer::~MyPlayer()
{
}

int MyPlayer::GetAmmo(void) const
{
	return ammo;
}

void MyPlayer::SetAmmo(int ammo_in)
{
	ammo = ammo_in;
}

int MyPlayer::AddAmmo(int ammo_in)
{
	ammo += ammo_in;
	return ammo;
}

int MyPlayer::UseAmmo(int& use)
{
	use = min(use, ammo);
	ammo -= use;
	return ammo;
}

std::string MyPlayer::GetScoreText(void) const
{
	std::stringstream str;

	str << "Ammo: ";
	if(ammo == 0)
		str << "--";
	else
		str << ammo;

	return str.str();
}


class MyHomeWorld : public Gaggle::HomeWorld
{
	public:
		MyHomeWorld(const std::string& name, unsigned short maxconnects, unsigned short homeport);
		~MyHomeWorld();

	protected:
		Gaggle::World* JoinPeerWorld(const Gaggle::WorldInfoMessage& msg, Gaggle::NetworkID peer_networkid);
		void LeftPeerWorld(Gaggle::World* world);

		Gaggle::Player* JoinPlayer(const Gaggle::JoinPlayerMessage& msg, Gaggle::NetworkID player_networkid);
		void EnterPlayer(Gaggle::Player& player);
		void PlayerHasLeft(Gaggle::Player* player);
		void PlayerAction(Gaggle::Player& player, const Gaggle::PlayerActionMessage& msg);

		void PlayerTexture(Gaggle::Player& player, const Gaggle::TextureMessage& msg);
		void PlayerModel(Gaggle::Player& player, const Gaggle::ModelMessage& msg);

		const Gaggle::Texture* AskTexture(Gaggle::Player& player, const Gaggle::AskTextureMessage& msg);
		const Gaggle::Model* AskModel(Gaggle::Player& player, const Gaggle::AskModelMessage& msg);
		Gaggle::ObjectsMessage* AskObjects(Gaggle::Player& player, const Gaggle::AskObjectsMessage& msg);

		void MyHomeWorld::PlayerTextMessage(Gaggle::Player& player, const Gaggle::TextMessage& msg);

		Gaggle::WorldLayoutMessage* AskWorldLayout(Gaggle::Player& player, const Gaggle::AskWorldLayoutMessage& msg);

	protected:
		// A simplistic way of creating UIDs. We'll just assign sequential
		// numbers. There are two things wrong with this approach:
		// 1) It may wrap around back to 0 and start to reuse already
		//    assigned UIDs.
		// 2) It does not avoid the special UNASSIGNED_UID value.
		Gaggle::UID next_object_uid;

		// A simplistic way of keeping the World Layout: just have a couple
		// of arrays with the World Points and World Squares. You probably
		// want to use something better for easier management.
		Gaggle::WorldPoint* worldpoints;
		Gaggle::WorldSquare* worldsquares;

		// Keep track of all the textures in the World
		std::list<Gaggle::Texture*> textures;
		// Should have something similar for Models
		// Keep track of all the Objects in the World
		std::list<Gaggle::Object*> objects;
		// May want to keep track of Players seperately
};


MyHomeWorld::MyHomeWorld(const std::string& name, unsigned short maxconnects, unsigned short homeport)
	: Gaggle::HomeWorld(name, 16, 16, 1, maxconnects, homeport)
{
	next_object_uid = 0;

	worldpoints = new Gaggle::WorldPoint[(numsquares[0] + 1) * (numsquares[1] + 1)];
	worldsquares = new Gaggle::WorldSquare[numsquares[0] * numsquares[1]];

	Gaggle::Texture *walltexture = new Gaggle::Texture(FILEPATH"/textures/mur079.jpg");
	walltexture->SetUID(next_object_uid++);
	textures.push_back(walltexture);

	// A trivial world layout, just a wall around a rectangular room
	// Could also be done by reading some data from a file or some
	// static data.
	Gaggle::WorldPoint* worldpoint = worldpoints;
	for(unsigned int y = 0; y < (numsquares[1]+1); y++)
	{
		for(unsigned int x = 0; x < (numsquares[0]+1); x++)
		{
			worldpoint->h[0] = 0;
			worldpoint->h[1] = 0.75f;
			worldpoint->h[2] = 1.50f;
			worldpoint->h[3] = 2.25f;
			worldpoint++;
		}
	}
	Gaggle::WorldSquare* worldsquare = worldsquares;
	for(unsigned int y = 0; y < numsquares[1]; y++)
	{
		for(unsigned int x = 0; x < numsquares[0]; x++)
		{
			// Start off with all open, then close walls on boundaries
			worldsquare->sectionsclosed[0] = 0;
			worldsquare->sectionsclosed[1] = 0;
			worldsquare->sectionsclosed[2] = 0;
			worldsquare->sectionsclosed[3] = 0;
			if(x == 0)						// Wall 3 entirely closed
				worldsquare->sectionsclosed[3] = 7;
			if(x == (numsquares[0]-1))	// Wall 1 entirely closed
				worldsquare->sectionsclosed[1] = 7;
			if(y == 0)						// Wall 2 entirely closed
				worldsquare->sectionsclosed[2] = 7;
			if(y == (numsquares[1]-1))	// Wall 0 entirely closed
				worldsquare->sectionsclosed[0] = 7;

			worldsquare->walltexture[0].texture = walltexture->GetUID();
			worldsquare->walltexture[0].scale = 1;
			worldsquare->walltexture[0].offset[0] = 0;
			worldsquare->walltexture[0].offset[1] = 0;
			worldsquare->walltexture[1].texture = walltexture->GetUID();
			worldsquare->walltexture[1].scale = 1;
			worldsquare->walltexture[1].offset[0] = 0;
			worldsquare->walltexture[1].offset[1] = 0;
			worldsquare->walltexture[2].texture = walltexture->GetUID();
			worldsquare->walltexture[2].scale = 1;
			worldsquare->walltexture[2].offset[0] = 0;
			worldsquare->walltexture[2].offset[1] = 0;
			worldsquare->walltexture[3].texture = walltexture->GetUID();
			worldsquare->walltexture[3].scale = 1;
			worldsquare->walltexture[3].offset[0] = 0;
			worldsquare->walltexture[3].offset[1] = 0;
			worldsquare->floortexture.texture = walltexture->GetUID();
			worldsquare->floortexture.scale = 2;
			worldsquare->floortexture.offset[0] = 0;
			worldsquare->floortexture.offset[1] = 0;
			worldsquare->ceilingtexture.texture = walltexture->GetUID();
			worldsquare->ceilingtexture.scale = 2;
			worldsquare->ceilingtexture.offset[0] = 0.5f;
			worldsquare->ceilingtexture.offset[1] = 0.5f;

			worldsquare->walllight[0] = 255;
			worldsquare->walllight[1] = 200;
			worldsquare->walllight[2] = 100;
			worldsquare->walllight[3] = 150;
			worldsquare->floorlight = 255;
			worldsquare->ceilinglight = 200;

			worldsquare++;
		}
	}
}


MyHomeWorld::~MyHomeWorld()
{
	for(std::list<Gaggle::Texture*>::iterator ti = textures.begin(); ti != textures.end(); ti++)
		delete *ti;

	for(std::list<Gaggle::Object*>::iterator oi = objects.begin(); oi != objects.end(); oi++)
		delete *oi;

	delete[] worldpoints;
	delete[] worldsquares;
}


// HomeWorld has joined a peer world
// Do any needed processing
Gaggle::World* MyHomeWorld::JoinPeerWorld(const Gaggle::WorldInfoMessage& msg, Gaggle::NetworkID peer_networkid)
{
	std::cerr << "Home World has joined peer" << std::endl;

	World *world = new Gaggle::World(msg, peer_networkid);

	std::cerr << "  Name " << world->GetName() << std::endl;

	// May want to add an Object to our World to represent the connection
	// to the peer World, which could for example touch to be transported
	// to that peer World. Remember to tell all the players about any new
	// objects.

	return world;
}


// HomeWorld has left a peer world, either because the HomeWorld
// disconnected, or because the peer World disconnected
// Do any needed processing
void MyHomeWorld::LeftPeerWorld(Gaggle::World* world)
{
	std::cerr << "Home World has left peer" << std::endl;

	delete world;
}


// HomeWorld is being joined by a Player. This is called as a
// result of a join request message being received
// Do any needed processing
Gaggle::Player* MyHomeWorld::JoinPlayer(const Gaggle::JoinPlayerMessage& msg, Gaggle::NetworkID player_networkid)
{
	std::cerr << "Home World has been joined by player" << std::endl;

	// Create a new player. This player object will be given to us
	// whenever something needs to be done with the Player.
	MyPlayer* player = new MyPlayer(msg, player_networkid);
	player->SetUID(next_object_uid++);

	// Start the Player in the middle of the World.
	// Should actually place the Player somewhere where there is
	// guaranteed to be space. Don't want Players to materialise
	// inside other Players for example.
	player->SetPos(numsquares[0]/2.f, 0, numsquares[1]/2.f);
	player->SetHeading(0);

	std::cerr << "  Name " << player->GetName() << std::endl;
	std::cerr << "  Given the player UID " << player->GetUID() << std::endl;

	// Gaggle will ask the new player for its model and texture

	return player;
}


// A Player that has joined is ready to start playing, and is
// about to be welcomed and introduced to the other Players.
// This is called when the Home World has all the needed info
// about the Player (like the Model and Texture), and the Player
// has signalled readiness to start playing.
// Do any needed processing
void MyHomeWorld::EnterPlayer(Gaggle::Player& player)
{
	MyPlayer& myplayer = (MyPlayer &)player;

	std::cerr << "Player UID " << myplayer.GetUID() << " has entered the home world" << std::endl;

	objects.push_back(&myplayer);

	// Give the Player some ammo
	myplayer.SetAmmo(30);

	Gaggle::TextMessage msg;
	std::string text;
	text = myplayer.GetName();
	text += " has entered the room";
	msg.Fill(msg, text);
	SendToPlayers(msg, myplayer.GetNetworkID());
}


// A Player has left the HomeWorld, either because the HomeWorld
// has disconnected it, or the Player has disconnected from the HomeWorld
// Do any needed processing
void MyHomeWorld::PlayerHasLeft(Gaggle::Player* player)
{
	std::cerr << "Player UID " << player->GetUID() << " has left the home world" << std::endl;

	// Remove the Player from our list of Objects in the World
	for(std::list<Gaggle::Object*>::iterator oi = objects.begin(); oi != objects.end(); oi++)
	{
		if(*oi == player)
		{
			objects.erase(oi);
			break;
		}
	}

	// We only use the Texture of the Player for the Player; so now
	// that the Player is gone, get rid of the Texture.
	for(std::list<Gaggle::Texture*>::iterator ti = textures.begin(); ti != textures.end(); ti++)
	{
		if((*ti)->GetUID() == player->GetTextureUID())
		{
			delete *ti;
			textures.erase(ti);
			break;
		}
	}

	delete player;
}


// A Player wants to do something in the HomeWorld
// Process the request and send an updated player state to all players
void MyHomeWorld::PlayerAction(Gaggle::Player& player, const Gaggle::PlayerActionMessage& msg)
{
	MyPlayer& myplayer = (MyPlayer &)player;

	std::cerr << "Player UID " << myplayer.GetUID() << " wants action " << msg.code << " value " << msg.value << std::endl;

	unsigned long curtime = RakNetGetTime();

	// Figure out what we think the Player's position and heading
	// is now
	Gaggle::Vector pos, dpos;
	myplayer.GetExtrapolatedPos(curtime, pos);
	myplayer.GetExtrapolatedDPos(curtime, dpos);
	float heading, dheading;
	heading = myplayer.GetExtrapolatedHeading(curtime);
	dheading = myplayer.GetDHeading();

	// As a simplistic example, we'll limit the position to 0,
	// and do only a simple linear movement. No gravity.
	// We'll take the Player's heading to be the forward direction.
	// Depending on what sort of game is being implemented, the
	// forward direction may be some other direction (e.g. +Z)

	pos.y = 0;
	dpos.y = 0;

	if(msg.code == Gaggle::ACTION_NONE)
		std::cerr << "  No action" << std::endl;
	else if(msg.code < 256)
		std::cerr << "  Key action" << std::endl;
	else if(msg.code == Gaggle::ACTION_MOVEFORWARD)
	{
		std::cerr << "  Move forward" << std::endl;
		dpos.x = (float)sin(heading) * msg.value;
		dpos.z = (float)cos(heading) * msg.value;
	}
	else if(msg.code == Gaggle::ACTION_MOVEBACKWARD)
	{
		std::cerr << "  Move backward" << std::endl;
		dpos.x = (float)-sin(heading) * msg.value;
		dpos.z = (float)-cos(heading) * msg.value;
	}
	else if(msg.code == Gaggle::ACTION_MOVELEFT)
	{
		std::cerr << "  Move left" << std::endl;
		dpos.x = (float)cos(heading) * msg.value;
		dpos.z = (float)-sin(heading) * msg.value;
	}
	else if(msg.code == Gaggle::ACTION_MOVERIGHT)
	{
		std::cerr << "  Move right" << std::endl;
		dpos.x = (float)-cos(heading) * msg.value;
		dpos.z = (float)sin(heading) * msg.value;
	}
	else if(msg.code == Gaggle::ACTION_TURNLEFT)
	{
		std::cerr << "  Turn left" << std::endl;
		dheading = msg.value;
	}
	else if(msg.code == Gaggle::ACTION_TURNRIGHT)
	{
		std::cerr << "  Turn right" << std::endl;
		dheading = -msg.value;
	}
	else if(msg.code == Gaggle::ACTION_JUMP)
	{
		std::cerr << "  Jump!" << std::endl;
	}
	else if(msg.code == Gaggle::ACTION_STOP)
	{
		std::cerr << "  Stop!" << std::endl;
		dpos.x = dpos.y = dpos.z = 0;
		dheading = 0;
	}
	else if(msg.code == Gaggle::ACTION_ACTIONBUTTON)
	{
		std::cerr << "  Action button" << std::endl;
		int useammo = 1;
		myplayer.UseAmmo(useammo);
		Gaggle::ScoreMessage scoremsg;
		scoremsg.Fill(scoremsg, myplayer.GetScoreText());
		SendToPlayer(scoremsg, myplayer);
	}

	// Store the new state.
	// We'll let the Player move forever. In an actual
	// game you may want to keep the Player moving until
	// it hits a wall or object by computing the time
	// until collision given the movement.
	player.SetPos(pos);
	player.SetDPos(dpos);
	player.SetDDPos(0,0,0);
	player.SetHeading(heading);
	player.SetDHeading(dheading);
	player.SetStartTime(curtime);
	player.SetEndTime(ULONG_MAX);
	// May also want to set the animation loop if the Player's
	// Model is an MD2.

	// Inform all Players (including this Player) about the
	// new state of the Object
	Gaggle::ObjectStateMessage statemsg;
	statemsg.Fill(statemsg, player);
	SendToPlayers(statemsg);
}


// A player is telling us about their texture
// Do any needed processing
void MyHomeWorld::PlayerTexture(Gaggle::Player& player, const Gaggle::TextureMessage& msg)
{
	std::cerr << "Player UID " << player.GetUID() << " has provided us with a texture:" << std::endl;
	std::cerr << "  compressed size " << msg.compressedsize << std::endl;

	// Remember the Player's Texture. We don't reuse Player textures,
	// so each Player's texture is considered unique.
	Gaggle::Texture* texture = new Gaggle::Texture(msg);
	texture->SetUID(next_object_uid++);
	textures.push_back(texture);
}


// A player is telling us about their model
// Do any needed processing
void MyHomeWorld::PlayerModel(Gaggle::Player& player, const Gaggle::ModelMessage& msg)
{
	std::cerr << "Player UID " << player.GetUID() << " has provided us with a model:" << std::endl;
	std::cerr << "  compressed size " << msg.compressedsize << std::endl;
	std::cerr << "  modeltype " << msg.modeltype << std::endl;

	// Should do something like with PlayerTexture()
}


// A player is asking us about a texture
// Do any needed processing, possibly returning a texture
const Gaggle::Texture* MyHomeWorld::AskTexture(Gaggle::Player& player, const Gaggle::AskTextureMessage& msg)
{
	std::cerr << "Got asked about texture UID " << msg.uid << std::endl;

	// Find the requested Texture in our list of Textures, and return it.
	for(std::list<Gaggle::Texture*>::iterator ti = textures.begin(); ti != textures.end(); ti++)
		if((*ti)->GetUID() == msg.uid)
			return *ti;

	return NULL;
}


// A player is asking us about a model
// Do any needed processing, possibly returning a model
const Gaggle::Model* MyHomeWorld::AskModel(Gaggle::Player& player, const Gaggle::AskModelMessage& msg)
{
	std::cerr << "Got asked about model UID " << msg.uid << std::endl;

	// Should do something like AskTexture()

	return NULL;
}


// A player is asking us about the objects in the World
// Do any needed processing, possibly sending an ID_OBJECTS_MESSAGE
Gaggle::ObjectsMessage* MyHomeWorld::AskObjects(Gaggle::Player& player, const Gaggle::AskObjectsMessage& msg)
{
	std::cerr << "Got asked about objects" << std::endl;

	// Create an ObjectMessage large enough for all the objects we have
	Gaggle::ObjectData* objectdata;
	Gaggle::ObjectsMessage* objectsmsg = Gaggle::ObjectsMessage::Create(objects.size(), objectdata);

	// Fill in the ObjectData part of the ObjectsMessage with the UIDs
	for(std::list<Gaggle::Object*>::iterator oi = objects.begin(); oi != objects.end(); oi++)
	{
		objectdata->uid = (*oi)->GetUID();
		objectdata->modeluid = (*oi)->GetModelUID();
		objectdata->textureuid = (*oi)->GetTextureUID();
		objectdata++;
	}

	return objectsmsg;
}


// A player has sent a us a text message
// Do any needed processing, for example broadcasting to all the players
void MyHomeWorld::PlayerTextMessage(Gaggle::Player& player, const Gaggle::TextMessage& msg)
{
	std::cerr << "Got text message:" << std::endl;
	std::cerr << "--" << std::endl;
	std::cerr << msg.text << std::endl;
	std::cerr << "--" << std::endl;

	// Broadcast the text message to all the Players
	// Maybe also have some special messages with which the
	// Player can affect the World, which don't get broadcast

	// Prepend player's name
	std::string text(player.GetName());
	text += ": ";
	text += msg.text;

	Gaggle::TextMessage textmsg;
	textmsg.Fill(textmsg, text);
	SendToPlayers(textmsg);
}


// A player is asking about part of the World layout
// Do any needed processing, possibly returning a WorldLayoutMessage
Gaggle::WorldLayoutMessage* MyHomeWorld::AskWorldLayout(Gaggle::Player& player, const Gaggle::AskWorldLayoutMessage& msg)
{
	std::cerr << "Player is asking about the layout in grid rectangle:" << std::endl;
	std::cerr << "  (" << msg.box[0][0] << ", " << msg.box[0][1] << ")";
	std::cerr << "-> (" << msg.box[1][0] << ", " << msg.box[1][1] << ")";

	// Return a rectangular region of the World Layout
	// The returned rectangle of grids does not need to be
	// the exact same rectangle as requested. At the very least
	// clip the rectangle to the extent of the World.
	// You could also limit the size of the rectangle to prevent
	// very large amounts of data being sent in one go.
	int x0 = max(msg.box[0][0], 0);
	int y0 = max(msg.box[0][1], 0);
	int x1 = min(msg.box[1][0], (int)numsquares[0]-1);
	int y1 = min(msg.box[1][1], (int)numsquares[1]-1);

	// Create a World Layout message with sufficient space for
	// the grid rectangle.
	Gaggle::WorldPoint* msgpoints;
	Gaggle::WorldSquare* msgsquares;
	Gaggle::WorldLayoutMessage *layoutmsg = Gaggle::WorldLayoutMessage::Create(x0, y0, x1, y1, msgpoints, msgsquares);

	// Creation could fail if the rectangle is invalid
	if(!layoutmsg)
		return NULL;

	// Copy over all the World Points of the grid squares in the rectangle
	for(int y = y0; y <= y1+1; y++)
	{
		for(int x = x0; x <= x1+1; x++)
		{
			*msgpoints = worldpoints[y * (numsquares[0]+1) + x];
			msgpoints++;
		}
	}
	// Copy over all the World Squares in the rectangle
	for(int y = y0; y <= y1; y++)
	{
		for(int x = x0; x <= x1; x++)
		{
			*msgsquares = worldsquares[y * numsquares[0] + x];
			msgsquares++;
		}
	}

	return layoutmsg;
}


//
// Test it
//


bool RunMyHomeWorld(int worldnum)
{
	int homeport = Gaggle::GAGGLE_PORT + worldnum;
	const char* worldnames[] = { "AlphaWorld", "BetaWorld", "GammaWorld", "DeltaWorld" };

	MyHomeWorld homeworld(worldnames[worldnum], 3, homeport);

	// Try connecting to another world
	if(worldnum != 0)
		if(!homeworld.Join("127.0.0.1", Gaggle::GAGGLE_PORT))
			return false;

	std::cerr << "Running my home world with name " << worldnames[worldnum] << " at port " << homeport << std::endl;

	while(1)
	{
		homeworld.ProcessPackets();

		unsigned char ch = 0;
		if(kbhit())
			ch = getch();

		if(ch == 'q')
			break;
	}

	return true;
}

bool RunMyHomePlayer(int playernum)
{
	const char* playernames[] = { "AlphaPlayer", "BetaPlayer", "GammaPlayer", "DeltaPlayer" };

	Gaggle::Texture texture(FILEPATH"/textures/hellpig.jpg");
	Gaggle::MD2Model model(FILEPATH"/models/hellpig.md2");

	MyHomePlayer homeplayer(playernames[playernum], &model, &texture);

	// Try connecting to a world
	if(!homeplayer.Join("127.0.0.1", Gaggle::GAGGLE_PORT))
		return false;

	std::cerr << "Running my home player with name " << playernames[playernum] << std::endl;

	Gaggle::Vector oldpos = { 0,0,0 };
	float oldheading = 0;
	unsigned long stoptime = 0;
	while(1)
	{
		homeplayer.ProcessPackets();

		unsigned char ch = 0;
		if(kbhit())
			ch = getch();

		if(ch == 'q')
			break;

		if(ch == 'a')
			homeplayer.Action(Gaggle::ACTION_TURNLEFT, (float)M_PI/2);
		else if(ch == 'A')
			homeplayer.Action(Gaggle::ACTION_MOVELEFT, 1);
		else if(ch == 'd')
			homeplayer.Action(Gaggle::ACTION_TURNRIGHT, (float)M_PI/2);
		else if(ch == 'D')
			homeplayer.Action(Gaggle::ACTION_MOVERIGHT, 1);
		else if(ch == 'w')
			homeplayer.Action(Gaggle::ACTION_MOVEFORWARD, 1);
		else if(ch == 'W')
			homeplayer.Action(Gaggle::ACTION_MOVEFORWARD, 4);
		else if(ch == 's')
			homeplayer.Action(Gaggle::ACTION_MOVEBACKWARD, 1);
		else if(ch == 'D')
			homeplayer.Action(Gaggle::ACTION_MOVEBACKWARD, 4);
		else if(ch == ' ')
			homeplayer.Action(Gaggle::ACTION_ACTIONBUTTON, 0);
		else if(ch == 13)	// 'Enter' key
			homeplayer.Action(Gaggle::ACTION_JUMP, 10);
		else if(ch > 0)	// 0 < ch < 256 are sent as ASCII keycodes
			homeplayer.Action((Gaggle::ActionCode)ch);

		unsigned long curtime = RakNetGetTime();
		unsigned long starttime = homeplayer.GetStartTime();
		// Force a stop after 1 second, as this simplistic example does
		// not have key up detection to stop moving.
		// What should happen when for example the user wants to stop
		// moving forward, is that an ACTION_MOVEFORWARDS action message
		// is sent with value set to 0.
		if(homeplayer.IsMoving() && (stoptime != starttime) &&
		   ((starttime + 1000) < curtime))
		{
			homeplayer.Action(Gaggle::ACTION_STOP);
			stoptime = starttime;
		}
		Gaggle::Vector pos;
		homeplayer.GetExtrapolatedPos(curtime, pos);
		float heading = homeplayer.GetExtrapolatedHeading(curtime);
		if((pos.x != oldpos.x) || (pos.y != oldpos.y) || (pos.z != oldpos.z) ||
		   (heading != oldheading))
		{
			std::cerr << "(" << pos.x << ", " << pos.y << ", " << pos.z << ")" << " " << heading << '\r';
			oldpos = pos;
			oldheading = heading;
		}
	}

	return true;
}


int main(int argc, char *argv[])
{
	std::cout << "Run as (W)orld or (P)layer?" << std::endl;
	char c;
	std::cin >> c;
	bool runworld;
	if((c == 'w') || (c == 'W'))
		runworld = true;
	else if((c == 'p') || (c == 'P'))
		runworld = false;
	else
		return EXIT_FAILURE;

	int num;
	std::cout << "Port offset (0, 1, 2, or 3)?" << std::endl;
	std::cin >> num;

	if((num < 0) || (num > 3))
		return EXIT_FAILURE;

	if(runworld)
		RunMyHomeWorld(num);
	else
		RunMyHomePlayer(num);

	return EXIT_SUCCESS;
}
